# -*- coding: utf-8 -*-
import sys
import time
import wave 
import pyaudio
import subprocess

from six.moves import queue
from google.api_core import exceptions
from google.cloud import speech_v1p1beta1 as speech
from google.cloud.speech_v1p1beta1 import enums
from google.cloud.speech_v1p1beta1 import types

class Speaker(object):

	def __init__(self):

		self.chunk_size = 1024
		self.p = pyaudio.PyAudio()

		## Bluetoothスピーカーの情報を取得
		for device_index in range(self.p.get_device_count()):
			device = self.p.get_device_info_by_index(device_index)
			if device["name"] == "bluealsa":
				self.bluetooth_device = device_index
				break

		## Bluetoothデバイスのチェック
		try:
			self.bluetooth_device


		except AttributeError:
			raise AttributeError("Bluetooth device not found.")
	

	def talk(self, filepath):

		try:
			wf = wave.open(filepath, "rb")

		except FileNotFoundError:
			raise ValueError("No such file " + filepath)

		try:

			## Bluetoothスピーカーを指定してストリームオープン
			self.stream = self.p.open(
				format=self.p.get_format_from_width(wf.getsampwidth()),
				channels=wf.getnchannels(),
				rate=wf.getframerate(),
				output=True,
				output_device_index=self.bluetooth_device
			)

		except OSError:
			raise OSError("Bluetooth device is busy.")

		data = wf.readframes(self.chunk_size)

		while len(data) > 0:
			self.stream.write(data)
			data = wf.readframes(self.chunk_size)

		self.stream.stop_stream()
		self.stream.close()


	def read(self, text):

		open_jtalk=["/usr/bin/open_jtalk"]
		mech=["-x", "/var/lib/mecab/dic/open-jtalk/naist-jdic"]
		htsvoice=['-m','/usr/share/hts-voice/mei/mei_happy.htsvoice']
		speed=['-r','1.0']
		outwav=['-ow','/tmp/open_jtalk.wav']
		cmd=open_jtalk+mech+htsvoice+speed+outwav

		c = subprocess.Popen(cmd, stdin=subprocess.PIPE)
		c.stdin.write(text.encode())
		c.stdin.close()
		c.wait()

		self.talk("/tmp/open_jtalk.wav")


	def __del__(self):
		self.p.terminate()


# Copyright 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MicrophoneStream(object):

	def __init__(self, rate, chunk):
		self._rate = rate
		self._chunk = chunk
		self._buff = queue.Queue()
		self.closed = True

	def __enter__(self):
		self._audio_interface = pyaudio.PyAudio()
		self._audio_stream = self._audio_interface.open(
			format=pyaudio.paInt16,
			channels=1, rate=self._rate,
			input=True, frames_per_buffer=self._chunk,
			stream_callback=self._fill_buffer
		)

		self.closed = False
		return self

	def __exit__(self, _type, value, traceback):
		self._audio_stream.stop_stream()
		self._audio_stream.close()
		self.closed = True
		self._buff.put(None)
		self._audio_interface.terminate()

	def _fill_buffer(self, in_data, _frame_count, _time_info, _status_flags):
		self._buff.put(in_data)
		return None, pyaudio.paContinue

	def generator(self):
		while not self.closed:
			chunk = self._buff.get()
			if chunk is None:
				return
			data = [chunk]

			while True:
				try:
					chunk = self._buff.get(block=False)
					if chunk is None:
						return

					data.append(chunk)

				except queue.Empty:
					break

			yield b''.join(data)


class GoogleCloudSpeech:

	def __init__(self, callbacks=None, console=True, rate=16000):

		if isinstance(callbacks, dict):
			for name in callbacks:

				if not callable(callbacks[name]):
					raise ValueError("Callback {} is not callable.".format(name))
					self.callbacks = callbacks

		else:
			self.callbacks = {}

		self.rate = rate
		self.console = console

	def listen_print_loop(self, responses):

		num_chars_printed = 0

		for response in responses:

			if not response.results:
				continue

			result = response.results[0]

			if not result.alternatives:
				continue

			## Speech-to-Text API レスポンス: https://cloud.google.com/speech-to-text/docs/basics?hl=ja
			transcript = result.alternatives[0].transcript
			confidence = result.alternatives[0].confidence

			overwrite_chars = " " * (num_chars_printed - len(transcript))

			if result.is_final:

				if confidence >= 0.7:
					text = transcript + overwrite_chars
					sout = "{} ({})\n".format(text, confidence)
					sys.stdout.write(sout)
					sys.stdout.flush()

				else:
					text = None
			

				self.callbacks.get("transcript", lambda x: True)(transcript)
				num_chars_printed = 0

				return text

	def listen(self, language_code='ja-JP'):

		client = speech.SpeechClient()

		config = types.RecognitionConfig(
			encoding=enums.RecognitionConfig.AudioEncoding.LINEAR16,
			sample_rate_hertz=self.rate,
			model=None,
			speech_contexts=[types.SpeechContext()],
			language_code=language_code
		)

		streaming_config = types.StreamingRecognitionConfig(
			config=config,
			single_utterance=True,
			interim_results=True
		)

		self.callbacks.get("ready", lambda: True)()

		with MicrophoneStream(self.rate, int(self.rate/10)) as stream:

			text = None
			self.callbacks.get("start", lambda: True)()

			try:
				audio_generator = stream.generator()
				requests = (types.StreamingRecognizeRequest(audio_content=content) for content in audio_generator)
				responses = client.streaming_recognize(streaming_config, requests)
				text = self.listen_print_loop(responses)

			except:
				print("Unexpected error:", sys.exc_info()[0])
				raise

			self.callbacks.get("end", lambda: True)()

			return text

	def on(self, name, callfunc):

		if callable(callfunc):
			self.callbacks[name] = callfunc
			return True

		return False

	def off(self, name):

		if name in self.callbacks:
			self.callbacks.pop(name)
			return True

		return False

